(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import AdomLib Util.
Require Import AbIdealEnv.

Instance ab_ideal_env_product {var: Type} {varEq:EqDec var} {A B:Type}
         `(@ab_ideal_env var varEq A) `(@ab_ideal_env var varEq B) : ab_ideal_env (A*B)%type :=
  {| id_top chan := let '(a, chan) := id_top chan in
                    let '(b, chan) := id_top chan in
                    ((a, b), chan)
   ; id_leb x y := let '((xa, xb), chan) := x in
                    let '(ya, yb) := y in
                    (id_leb (xa, chan) ya && id_leb (xb, chan) yb)%bool
   ; id_join x y chan := let '((xa, xb), xchan) := x in
                         let '((ya, yb), ychan) := y in
                         let '(reta, chan) := id_join (xa, xchan) (ya, ychan) chan in
                         let '(retb, chan) := id_join (xb, xchan) (yb, ychan) chan in
                         ((reta, retb), chan)
   ; id_widen x y chan := let '((xa, xb), xchan) := x in
                          let '((ya, yb), ychan) := y in
                          let '(reta, chan) := id_widen (xa, xchan) (ya, ychan) chan in
                          let '(retb, chan) := id_widen (xb, xchan) (yb, ychan) chan in
                          ((reta, retb), chan)
   ; assign v e ab chan := let '(a, b, abchan) := ab in
                           do_bot retachan <- assign v e (a, abchan) chan;
                           let '(reta, chan) := retachan in
                           do_bot retbchan <- assign v e (b, abchan) chan;
                           let '(retb, chan) := retbchan in
                           NotBot ((reta, retb), chan)
   ; forget v ab chan := let '(a, b, abchan) := ab in
                         do_bot retachan <- forget v (a, abchan) chan;
                         let '(reta, chan) := retachan in
                         do_bot retbchan <- forget v (b, abchan) chan;
                         let '(retb, chan) := retbchan in
                         NotBot ((reta, retb), chan)
   ; assume e bo ab chan := let '(a, b, abchan) := ab in
                            do_bot retachan <- assume e bo (a, abchan) chan;
                            let '(reta, chan) := retachan in
                            do_bot retbchan <- assume e bo (b, abchan) chan;
                            let '(retb, chan) := retbchan in
                            NotBot ((reta, retb), chan)
  |}.
Proof.
  - intros.
    pose proof (id_top_correct (t:=A) H1). destruct (id_top chan), H2.
    pose proof (id_top_correct (t:=B) H3). simpl in H4. destruct (id_top i), H4.
    constructor. constructor. auto. auto. auto.
  - intros. destruct ab1 as [[] ?], ab2.
    rewrite Bool.andb_true_iff in H1. destruct H1.
    destruct H2. destruct H2.
    constructor. eapply id_leb_correct in H1. eauto. constructor. auto. auto.
    eapply id_leb_correct in H3. eauto. constructor. auto. auto.
  - intros. destruct ab1 as [[] ?], ab2  as [[] ?].
    pose proof (id_join_correct (t:=A) (ab1:=(a, i)) (ab2:=(a0, i0)) (chan:=chan)).
    edestruct (id_join (a, i) (a0, i0) chan), H3.
    destruct H1 as [[[] ?]|[[] ?]]; [left|right]; repeat (constructor; eauto).
    auto.
    pose proof (id_join_correct (t:=B) (ab1:=(b, i)) (ab2:=(b0, i0)) (chan:=i1)).
    edestruct (id_join (b, i) (b0, i0) i1), H6.
    destruct H1 as [[[] ?]|[[] ?]]; [left|right]; repeat (constructor; eauto).
    auto.
    repeat (constructor; auto).
  - intros. destruct ab as [[] ?]. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a0. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a1. destruct H4, H5. constructor. constructor. eauto. eauto. auto.
    apply assign_correct. destruct H1 as [[] ?]. constructor. auto. auto. destruct H4. auto. auto.
    apply assign_correct. destruct H1 as [[] ?]. constructor. auto. auto. auto. auto.
  - intros. destruct ab as [[] ?]. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a0. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a1. destruct H3, H4. constructor. constructor. eauto. eauto. auto.
    apply forget_correct. destruct H1 as [[] ?]. constructor. auto. auto. destruct H3. auto.
    apply forget_correct. destruct H1 as [[] ?]. constructor. auto. auto. auto.
  - intros. destruct ab as [[] ?]. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a0. eapply (botbind_spec (GA:=WProd.gamma _ _)).
    intros. destruct a1. destruct H4, H5. constructor. constructor. eauto. eauto. auto.
    apply assume_correct. destruct H1 as [[] ?]. constructor. auto. auto. destruct H4. auto. auto.
    apply assume_correct. destruct H1 as [[] ?]. constructor. auto. auto. auto. auto.
  - intros chan a [? ?].
    pose proof (id_top_redundant (t:=A) (chan:=chan)). destruct (id_top chan).
    pose proof (id_top_redundant (t:=B) (chan:=i)). destruct (id_top i).
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
  - intros [[]?] [[]?] chan a [? ?].
    pose proof (id_join_redundant (a0, i) (a1, i0) (chan:=chan)).
    destruct (id_join (a0, i) (a1, i0) chan).
    pose proof (id_join_redundant (b, i) (b0, i0) (chan:=i1)).
    destruct (id_join (b, i) (b0, i0) i1).
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
  - intros [[]?] [[]?] chan a [? ?].
    pose proof (id_widen_redundant (a0, i) (a1, i0) (chan:=chan)).
    destruct (id_widen (a0, i) (a1, i0) chan).
    pose proof (id_widen_redundant (b, i) (b0, i0) (chan:=i1)).
    destruct (id_widen (b, i) (b0, i0) i1).
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
  - intros x e [[]?] chan a [? ?].
    pose proof (assign_redundant x e (a0, i) (chan:=chan)).
    destruct (assign x e (a0, i) chan) as [|[]]. contradiction. simpl. simpl in H2.
    pose proof (assign_redundant x e (b, i) (chan:=i0)).
    destruct (assign x e (b, i) i0) as [|[]]. contradiction.
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
  - intros x [[]?] chan a [? ?].
    pose proof (forget_redundant x (a0, i) (chan:=chan)).
    destruct (forget x (a0, i) chan) as [|[]]. contradiction. simpl. simpl in H2.
    pose proof (forget_redundant x (b, i) (chan:=i0)).
    destruct (forget x (b, i) i0) as [|[]]. contradiction.
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
  - intros e bo [[]?] chan a [? ?].
    pose proof (assume_redundant e bo (a0, i) (chan:=chan)).
    destruct (assume e bo (a0, i) chan) as [|[]]. contradiction. simpl. simpl in H2.
    pose proof (assume_redundant e bo (b, i) (chan:=i0)).
    destruct (assume e bo (b, i) i0) as [|[]]. contradiction.
    apply H4. split. apply H3. split. auto. destruct H2. auto. destruct H2. auto.
Defined.
