(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8 Util AdomLib.
Require Import Floats IdealExpr IdealIntervals.

Set Implicit Arguments.

(** Signature of an abstract numerical domain. *)
Unset Elimination Schemes.
Class ab_ideal_nonrel (t:Type) : Type :=
{ as_pl :> pre_lattice t
; as_gamma :> gamma_op t ideal_num

; meet: t+⊤ → t+⊤ → t+⊤+⊥

; z_const: ZIntervals.itv → t+⊤
; float_const: FloatIntervals.itv → t+⊤

; forward_unop: i_unary_operation → t+⊤ → t+⊤+⊥
; forward_binop: i_binary_operation → t+⊤ → t+⊤ → t+⊤+⊥

; backward_unop: i_unary_operation → t+⊤ → t+⊤ → t+⊤+⊥
; backward_binop: i_binary_operation → t+⊤ → t+⊤ → t+⊤ → (t+⊤*t+⊤)+⊥

; forward_try_subst: t+⊤ -> bool
; backward_try_subst: t+⊤ -> t+⊤ -> bool
}.

(** Specification of an abstract numerical domain. *)
Existing Instances WProd.make_wl WTensor.gamma.
Class ab_ideal_nonrel_correct t (D: ab_ideal_nonrel t) : Prop :=
{ as_pls :> pre_lattice_spec as_pl as_gamma
; meet_correct: ∀ x y, (γ x) ∩ (γ y) ⊆ γ (meet x y)
; z_const_correct: ∀ (i: ZIntervals.itv), γ i ⊆ (γ (z_const i) ∘ INz)
; float_const_correct: ∀ (f:FloatIntervals.itv), γ f ⊆ (γ (float_const f) ∘ INf)
; forward_unop_sound: ∀ op x x_ab, x ∈ γ x_ab ->
    eval_iunop op x ⊆ γ (forward_unop op x_ab)
; forward_binop_sound: ∀ op x x_ab, x ∈ γ x_ab -> ∀ y y_ab, y ∈ γ y_ab ->
    eval_ibinop op x y ⊆ γ (forward_binop op x_ab y_ab)

; backward_unop_sound:
    ∀ op x x_ab, x ∈ γ x_ab -> ∀ res res_ab, res ∈ γ res_ab ->
      eval_iunop op x res ->
      x ∈ γ (backward_unop op res_ab x_ab)

; backward_binop_sound:
    ∀ op x x_ab, x ∈ γ x_ab ->
    ∀ y y_ab, y ∈ γ y_ab ->
    ∀ res res_ab, res ∈ γ res_ab ->
      eval_ibinop op x y res ->
      (x, y) ∈ γ (backward_binop op res_ab x_ab y_ab)
}.
