(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Values.
Require Import Events.
Require Import Coqlib.
Require Import AST.
Require Import Globalenvs.
Require Import Utf8.
Require Import String.

Definition any_int64_ident := ident_of_string "any_int64"%string.

Definition any_int64_sem : extcall_sem :=
  fun F V ge args mem tr ret mem' =>
    exists r,
      args = nil /\
      mem' = mem /\
      tr = Event_syscall any_int64_ident nil (EVlong r)::nil /\
      ret = Vlong r.

Definition any_int64_sig :=
  {| sig_args := nil; sig_res := Some Tlong; sig_cc := cc_default |}.

Lemma any_int64_spec :
  extcall_properties any_int64_sem any_int64_sig.
Proof.
  constructor; intros.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor.
  - destruct H1 as (r & ? & ? & ? & ?); subst. repeat econstructor.
  - destruct H as (r & ? & ? & ? & ?); subst. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor. intuition. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. inv H1. repeat (econstructor; auto).
  - destruct H0 as (r & ? & ? & ? & ?); subst. inv H2.
    do 3 eexists. split. econstructor; eauto.
    split. constructor. split. eauto.
    split. constructor. easy. auto. split. constructor. easy. auto.
    split. intro. auto. intro. congruence.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor.
  - destruct H as (r & ? & ? & ? & ?); subst.
    inv H0. destruct res2; try discriminate.
    repeat econstructor; eauto.
  - destruct H as (r & ? & ? & ? & ?); subst.
    destruct H0 as (r' & ? & ? & ? & ?); subst.
    repeat constructor. inv H0. auto.
Qed.

(** We state that the semantics of any_int64 is what we think.
    This axiom is not contradictory, as we just impose one value of
    the parameter external_functions_sem to be any_int64_sem, and this
    value respects the property any_int64_spec. *)
Axiom any_int64_sem_eq :
  external_functions_sem any_int64_ident any_int64_sig = any_int64_sem.

Definition any_double_ident := ident_of_string "any_double"%string.

Definition any_double_sem : extcall_sem :=
  fun F V ge args mem tr ret mem' =>
    exists r,
      args = nil /\
      mem' = mem /\
      tr = Event_syscall any_double_ident nil (EVfloat r)::nil /\
      ret = Vfloat r.

Definition any_double_sig :=
  {| sig_args := nil; sig_res := Some Tfloat; sig_cc := cc_default |}.

Lemma any_double_spec :
  extcall_properties any_double_sem any_double_sig.
Proof.
  constructor; intros.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor.
  - destruct H1 as (r & ? & ? & ? & ?); subst. repeat econstructor.
  - destruct H as (r & ? & ? & ? & ?); subst. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor. intuition. auto.
  - destruct H as (r & ? & ? & ? & ?); subst. inv H1. repeat (econstructor; auto).
  - destruct H0 as (r & ? & ? & ? & ?); subst. inv H2.
    do 3 eexists. split. econstructor; eauto.
    split. constructor. split. eauto.
    split. constructor. easy. auto. split. constructor. easy. auto.
    split. intro. auto. intro. congruence.
  - destruct H as (r & ? & ? & ? & ?); subst. constructor.
  - destruct H as (r & ? & ? & ? & ?); subst.
    inv H0. destruct res2; try discriminate.
    repeat econstructor; eauto.
  - destruct H as (r & ? & ? & ? & ?); subst.
    destruct H0 as (r' & ? & ? & ? & ?); subst.
    repeat constructor. inv H0. auto.
Qed.

(** We state that the semantics of any_double is what we think.
    This axiom is not contradictory, as we just impose one value of
    the parameter external_functions_sem to be any_double_sem, and this
    value respects the property any_int64_spec. *)
Axiom any_double_sem_eq :
  external_functions_sem any_double_ident any_double_sig = any_double_sem.