(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8
  Coqlib
  Globalenvs
  Csharpminor
  Behaviors
  String.

Require Import
  Utils
  ShareTree
  ToString
  AbMachineEnv
  IdealEnvToMachineEnv
  IdealIntervalsNonrel
  PolAdomZ
  BlockMon
  AbMemSignatureCsharpminor
  DebugAbMemCsharpminor
  CsharpminorIter
  CsharpminorIterproof
  Cells MemCsharpminor
  AbIdealEnvProduct
  AbVarExpEq
  Hash.

Require Import
  DebugAbMachineEnv
  ToString
  AdomLib
  ZCongruences
  CellEncoding.

Section S.

  Context (max_concretize : N)
          (trace verbose: bool)
          (unroll: nat).

  Context {tnum: Type}
          `{ToString tnum}
          (M: @ab_machine_env cell tnum _).

  Existing Instance TToString.
  Definition doit (p: program) : block_mon unit :=
    let ge := Genv.globalenv p in
    let mem_dom := mem_cshm_dom M ge max_concretize in
    iter_program _ unroll p (if trace then debug_mem_dom mem_dom verbose
                             else mem_dom).

End S.

Local Unset Elimination Schemes.
Inductive num_dom_kind :=
| ZIntervals
| Poly
.

Module IETME_AC := IdealEnvToMachineEnv.Make ACTree.
Module ZCIE_AC := ZCongruences.Make ACTree.
Module IIE_AC := IdealIntervalsNonrel.Make ACTree.
Module VEEQ_AC := AbVarExpEq.Make ACTree.

Module IETME_P := IdealEnvToMachineEnv.Make PShareTree.
Module ZCIE_P := ZCongruences.Make PShareTree.
Module IIE_P := IdealIntervalsNonrel.Make PShareTree.
Module VEEQ_P := AbVarExpEq.Make PShareTree.

Module PTreeDom := WPFun PShareTree.

Existing Instance ACTreeDom.toString.
Existing Instance PTreeDom.toString.

Section VA.

  Context (kind: num_dom_kind)
          (max_concretize: N)
          (trace verbose: bool)
          (unroll: nat).

  Local Notation analysis :=
    (λ e u p,
     doit max_concretize trace verbose unroll
          (if trace
           then debug_ab_machine_env verbose e
           else e)
          p).


  Existing Instance CellAsPos.
  Definition vanalysis (p: program) : block_mon unit :=
    match kind with
    | Poly =>
      let dom := IETME_P.idealenv_to_machineenv
          (ab_ideal_env_product
             (VEEQ_P.abVarExpEq_env (fun h p => MIX_N h (Npos p)))
             (ab_ideal_env_product
                IIE_P.intervals_ideal_env
                (ab_ideal_env_product
                   ZCIE_P.zcongr_ideal_env
                   PolAbIdealEnv)))
      in
      analysis (wrap dom _) unroll p
    | _ =>
       let dom :=
           IETME_AC.idealenv_to_machineenv
             (ab_ideal_env_product
                (VEEQ_AC.abVarExpEq_env MIX_cell)
             (ab_ideal_env_product
                IIE_AC.intervals_ideal_env
                ZCIE_AC.zcongr_ideal_env))
       in
       analysis dom unroll p
    end.

  Theorem vanalysis_correct :
    forall prog, vanalysis prog = (tt, nil) ->
    forall tr, program_behaves (semantics prog) (Goes_wrong tr) ->
    False.
  Proof.
    unfold vanalysis, doit. intros.
    destruct kind, trace, verbose; eapply iter_program_sound; eauto;
    match goal with
      | |- mem_dom_spec _ (debug_mem_dom _ _) _ => eapply debug_mem_dom_sound
      | _ => idtac
    end;
    eapply mem_cshm_dom_correct.
  Qed.

End VA.
