(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String ToString AdomLib.

Parameter print_string : string → unit.

Definition print0 {A} : string → A → A*unit :=
  fun s x =>
  (x, print_string s).

Extraction NoInline print0.

Definition print {A} : string → A → A :=
  fun s x => fst (print0 s x).

Lemma print_id {A} s (x: A) : print s x = x.
Proof. now unfold print; destruct print_string. Qed.

Local Open Scope string_scope.

Section ADOM.

Context {t t': Type} WL G (D: adom t t' WL G) `{ToString t}.


Definition print_leb (x y: string) (r: bool) : string :=
  x ++ " ⊑ " ++ y ++ " → " ++ (if r then "tt" else "ff").

Definition print_join (x y z: string) : string :=
  x ++ " ⊔ " ++ y ++ " → " ++ z.

Definition print_widen (x y z: string) : string :=
  x ++ " ∇ " ++ y ++ " → " ++ z.

Instance debug_wlat : weak_lattice t :=
  { leb x y := let r := leb x y in
               let s := print_leb (to_string x) (to_string y) r in
               print s r
  ; top := top
  ; join x y := let r := join x y in
                let s := print_join (to_string x) (to_string y) (to_string r) in
                print s r
  ; widen x y := let r := widen x y in
                 let s := print_widen (to_string x) (to_string y) (to_string r) in
                 print s r
  }.

Lemma debug_adom : adom t t' debug_wlat G.
Proof.
  split.
  intros x y; simpl; rewrite print_id. apply gamma_monotone; auto.
  simpl. apply gamma_top; auto.
  intros x y; simpl; rewrite print_id. apply join_sound; auto.
Qed.

End ADOM.
